# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse
import ast
import logging
import os
import yaml

from box import Box

from mercury.common.exceptions import MercuryConfigurationError
from mercury.common.log import log_format


LOG = logging.getLogger(__name__)

DEFAULT_SEARCH_DIRS = ['.', '~/.mercury', '/etc/mercury']


def find_config(filename, dirs=None):
    """Searches for configuration files. Setting the environment variable:
        ```
            MERCURY_SEARCH_DIRS
        ```
        can be used to extend the search paths as well

    :param filename: The configuration file path or a filename to search for
    :param dirs: A list of filesystem directories to search for the
        configuration file. Parameter defaults to None, which means a set
        of default locations will be searched.
    :returns: string or None -- The full path of the configuration file if
        found.  None otherwise.
    """
    if os.path.isfile(filename):
        return filename

    dirs = dirs or list() + DEFAULT_SEARCH_DIRS
    if 'MERCURY_SEARCH_DIRS' in os.environ:
        dirs = dirs + os.environ.get('MERCURY_SEARCH_DIRS')

    for directory in dirs:
        full_path = os.path.join(os.path.expanduser(directory), filename)
        if os.path.isfile(full_path):
            return full_path


def configuration_from_yaml(filename):
    """Loads a YAML configuration file.

    :param filename: The filename of the file to load.
    :returns: dict -- A dictionary representing the YAML configuration file
        loaded. If the file can't be loaded, then the empty dict is returned.
    """
    try:
        with open(filename) as infile:
            return yaml.load(infile.read())
    except IOError:
        return {}


def get_configuration(filename):
    """Gets and returns the contents of the configuration file as a dict.

    :param filename: The filename of the configuration file to load.
    :returns: dict -- A dictionary representing the YAML configuration file
        loaded. If the file was not found or loaded, then returns an empty
        dict.
    """
    config_file = find_config(filename)
    if not config_file:
        LOG.warning('%s not found' % filename)
        return {}
    return configuration_from_yaml(config_file)


class MercuryConfiguration(object):
    def __init__(self,
                 program_name,
                 configuration_file,
                 description=None,
                 config_search_dirs=None):
        """Mercury Configuration helper which simplifies cli arguments,
        environment variables, and configuration file merging

        :param program_name:
        :param configuration_file: The name of the configuration file
        :param description:
        """
        self.program_name = program_name
        self.argument_parser = argparse.ArgumentParser(
            prog=self.program_name,
            description=description,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
        self.config_search_dirs = config_search_dirs or []
        self.master_configuration = Box()
        self.argparse_namespace = None
        self.options = []
        self.configuration_file = configuration_file
        self.configuration = {}

        self.argument_parser.add_argument('-c',
                                          '--configuration',
                                          default=self.configuration_file,
                                          help='Path to the configuration file')

        # Add logging at the common level, every program will need this
        # advanced logging facilities must be implemented at the program
        # level

        self.add_option('logging.level',
                        '--log-level',
                        'MERCURY_LOG_LEVEL',
                        'logging.level',
                        default='INFO',
                        help_string='Log level, INFO | DEBUG | ERROR | WARNING')

        self.add_option('logging.format',
                        '--log-format',
                        'MERCURY_LOG_FORMAT',
                        'logging.format',
                        default=log_format,
                        help_string='Desired logging format')

    def set_configuration(self):
        """ Parse the configuration """

        # self.configuration file is the default option
        if self.argparse_namespace.configuration != self.configuration_file:
            # Configuration file is explicitly defined on the command line
            self.configuration = configuration_from_yaml(
                self.argparse_namespace.configuration)
        else:
            # Search for the configuration file in the standard location
            config_file = find_config(self.configuration_file,
                                      self.config_search_dirs)
            if config_file:
                self.configuration = configuration_from_yaml(config_file)

        # Seed namespace with configuration to capture config only values
        self.master_configuration.update(self.configuration)

    @staticmethod
    def get_by_namespace(dictionary, namespace):
        """
        Used to access a dictionary values using dot notation

        For instance:

        .. code-block::

            {'one': {'two': 'three'}}

            >> d['one']['two']
            'three'
            >> get_by_namespace('one.two')
            'three'
        :param dictionary: The dictionary we are searching
        :param namespace: Dot notated namespace specifier
        :type namespace: str
        :return: The value the namespace resolves to or None
        """

        # creating copy of reference for readability
        value = dictionary
        for k in namespace.split('.'):
            try:
                value = value[k]
            except KeyError:
                return
        return value

    @staticmethod
    def set_by_namespace(dictionary, namespace, value):
        """
        Like get_by_namespace, but set a value rather than returning it
        :param dictionary: The dictionary we are modifying
        :param namespace:
        :param value:
        :return:
        """
        expanded = namespace.split('.')
        # Since the last element is the value reference, we iterate only over
        # [:-1] elements
        for k in expanded[:-1]:
            dictionary = dictionary[k]
        dictionary[expanded[-1]] = value

    @staticmethod
    def convert_argparse_destination(argument_name):
        """ Converts underscores to dashes, for a deterministic mechanism
         to name (and recall) argparse destinations"""
        return argument_name.lstrip('-').replace('-', '_')

    def parse_args(self):
        """ Add option arguments and parse_args """
        for opt in self.options:
            if not opt['added_to_parser']:
                self.argument_parser.add_argument(
                    opt['cli_argument'],
                    dest=self.convert_argparse_destination(opt['cli_argument']),
                    help=opt['help'],
                    *opt['args'],
                    **opt['kwargs'])
                # added_to_parser is here for instances where
                # populate_namespace is called more than once per class
                # instance. Such a thing happens in our test cases, and thus
                # could happen during implementation
                opt['added_to_parser'] = True

        self.argparse_namespace = self.argument_parser.parse_args()

    @staticmethod
    def format_type(value, special_type):
        """
        Since argparse (by default) and os.environ can only be strings, we allow
        options to specify a type
        :param value: The value we are transforming
        :param special_type: The 'type' we are transforming to
        :type special_type: list, int, float, bool
        :return: None
        """
        if type(special_type) is not type:
            raise MercuryConfigurationError(
                f'{special_type} must be a type, not an object'
            )
        if special_type not in [list, int, float, bool]:
            raise MercuryConfigurationError(
                f'{special_type} is not support for conversion')

        if isinstance(value, special_type):
            # Type is already converted
            return value

        if not isinstance(value, str):
            raise MercuryConfigurationError(
                'Only string types are converted'
            )

        if special_type is list:
            # Expected comma separated value, escaping is not supported
            return value.split(',')

        elif special_type is bool:
            # if the string value is 0 or FALSE (any case) then the value is
            # converted to False, otherwise True
            return value.lower() not in ['0', 'false']
        else:
            try:
                return special_type(value)
            except ValueError:
                raise MercuryConfigurationError(
                    f'{value} cannot be converted to '
                    f'{special_type}'
                )

    @staticmethod
    def expand_option(namespace, option):
        """ Expand an option, usually expressed in dotted notation, to a
        namespace path. All but the last element, the value element, are
        considered containers. For the option: nest1.nest2.nest3, nest1 and
        nest2 would be initialized as new boxes, if they did not exist already.
        nest3 would be initialized with the default value of the option
        :param namespace: self.namespace reference
        :param option: The option we are expanding
        """
        expanded_name = option['name'].split('.')
        for container in expanded_name[:-1]:
            if container not in namespace:
                namespace[container] = Box()
            # Check to see if the existing element is actually a container
            elif not isinstance(namespace[container], dict):
                raise MercuryConfigurationError(
                    f'{container} exists and is not a container, '
                    f'{option["name"]} is invalid')
            namespace = namespace[container]
        namespace[expanded_name[-1]] = option['default']

    def override_value(self, name, value, special_type):
        """
        DRY override method replacing a value from a given source
        :param name: The name of the value
        :param value: The value
        :param special_type: type to convert the value to
        """
        if value:
            if special_type:
                value = self.format_type(value, special_type)
            self.set_by_namespace(self.master_configuration, name, value)

    def override_with_configuration(self, option):
        """
        Overrides option with a value in the configuration file
        :param option: an option from self.options
        """
        self.override_value(option['name'],
                            self.get_by_namespace(self.configuration,
                                                  option['config_address']),
                            option['special_type'])

    def override_with_environment(self, option):
        """
        Overrides option with the value found in an environment variable
        :param option: an option from self.options
        :return:
        """
        self.override_value(option['name'],
                            os.environ.get(option['env_variable']),
                            option['special_type'])

    def override_with_cli_options(self, option):
        """
        Override option with a value passed in as a command line argument
        :param option: an option from self.options
        """
        self.override_value(option['name'],
                            vars(
                                self.argparse_namespace).get(
                                self.convert_argparse_destination(
                                    option['cli_argument'])),
                            option['special_type'])

    def scan_options(self):
        """
        :return: The master configuration
        """
        self.parse_args()
        self.set_configuration()

        for option in self.options:
            # populate the namespace and fill defaults
            self.expand_option(self.master_configuration, option)

            # override defaults with configuration values
            self.override_with_configuration(option)

            # override defaults and config options with environment variables
            self.override_with_environment(option)

            # command line
            self.override_with_cli_options(option)

            # validate
            value = self.get_by_namespace(self.master_configuration,
                                          option['name'])
            # Check required
            if option['required'] and not value:
                raise MercuryConfigurationError(f'Option {option["name"]} has '
                                                f'not been provided')
            if option['one_of'] and not value not in 'one_of':
                raise MercuryConfigurationError(
                    f'Option {option["name"]} is {value} but must be one of '
                    f'{",".join(str(option["one_of"]))}'
                )
        return self.master_configuration

    @staticmethod
    def check_name(name):
        """
        Validator of the argument name, enforces python namespace rules

        :param name: The option name
        :return: None
        :raises: MercuryConfigurationError
        """
        try:
            ast.parse(name)
        except SyntaxError:
            raise MercuryConfigurationError(f'{name} is not a valid name')

    @staticmethod
    def auto_format_cli_argument(name):
        """ It is a popular convention for arguments to have dashes rather than
        underscores, they should also be prepended with dashes

        .. note::

            Switches (-s) are not supported

        :param name: The option name
        :return: The formatted name
        """
        return '--' + name.replace('.', '-').replace('_', '-')

    @staticmethod
    def auto_format_env_variable(name):
        """ It is convention that environment variable use underscores and be
        upper case

        :param name: The option name.
        :return: The formatted name
        """
        return name.replace('.', '_').upper()

    def add_option(self,
                   name,
                   cli_argument=None,
                   env_variable=None,
                   config_address=None,
                   default=None,
                   help_string=None,
                   special_type=None,
                   required=False,
                   one_of=None,
                   *args,
                   **kwargs):
        """
        Allows us to define options which can be defined on the command line,
        from an environment variable, or from the configuration file.

        :param name: The option namespace
        :param cli_argument: The name argument name
        :param env_variable: The environment variable
        :param config_address: The 'address' of the option's configuration key.
         Configuration files are yaml. For the configuration:

        .. code-block: yaml

            nest1:
              nest2:
                nest3: My Value

        nest1.nest2.nest3 == 'My Value'

        :param default: The default value of the option
        :param help_string: Description of what the option does and how to use
            it
        :param special_type: Allows us to specify a type conversion for the
            value. Supported type conversions are int, float, and list
        :param required: An exception will be raised if the namespace is not
        populated by some source
        :param one_of:
        :param args: Additional args to pass to argparse
        :param kwargs: Additional kwargs to pass to argparse
        :return: None
        """
        self.check_name(name)
        self.options.append(dict(
            name=name,
            cli_argument=cli_argument or self.auto_format_cli_argument(name),
            env_variable=env_variable or self.auto_format_env_variable(name),
            config_address=config_address or name,
            default=default,
            help=help_string,
            special_type=special_type,
            required=required,
            one_of=one_of or [],
            args=args,
            kwargs=kwargs,
            added_to_parser=False,
        ))



"""
async_endpoints is a singleton, allowing dispatch endpoints to be
specified using a decorator. Controllers can define endpoints any way
they see fit. This is here for convenience only.

For an example usage, see mercury.inventory.controller.InventoryController
"""

import inspect
import logging

from mercury.common.exceptions import MercuryClientException

log = logging.getLogger(__name__)


async_endpoints = dict()


def async_endpoint(name):
    def add(f):
        log.debug('Adding async runtime endpoint {} ({})'.format(f.__name__, name))
        if not inspect.iscoroutinefunction(f):
            log.error('{} is not a coroutine'.format(f.__name__))
        elif name in async_endpoints:
            log.error('{} already exists in table'.format(name))
        else:
            async_endpoints[name] = f
        return f
    return add


class StaticEndpointController(object):
    """Use this if you want to be lazy"""
    def __init__(self):
        self.endpoints = async_endpoints

    # TODO: CODE Duplication, move these static methods to a library module

    @staticmethod
    def get_key(key, d):
        try:
            return d[key]
        except KeyError:
            raise MercuryClientException('{} is missing from request'.format(key))

    @staticmethod
    def validate_required(required, data):
        missing = []
        for key in required:
            if key not in data:
                missing.append(key)

        if missing:
            raise MercuryClientException('Message is missing required data: {}'.format(missing))



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

from mercury.common.clients.router_req_client import RouterReqClient

log = logging.getLogger(__name__)


class BackEndClient(RouterReqClient):
    service_name = 'RPC backend'

    def register(self, device_info, agent_info):
        """

        :param device_info:
        :param agent_info:
        :return:
        """
        _payload = {
            'endpoint': 'register',
            'args': [device_info, agent_info]
        }
        return self.transceiver(_payload)

    def update(self, mercury_id, update_data):
        """

        :param mercury_id:
        :param update_data:
        :return:
        """
        _payload = {
            'endpoint': 'update',
            'args': [mercury_id, update_data]
        }
        return self.transceiver(_payload)

    def complete_task(self, return_data):
        """
        :param return_data:
        :return:
        """
        _payload = {
            'endpoint': 'complete_task',
            'args': [return_data]
        }
        return self.transceiver(_payload)

    def update_task(self, update_data):
        """
        :param update_data:
        :return:
        """
        _payload = {
            'endpoint': 'update_task',
            'args': [update_data]
        }
        return self.transceiver(_payload)



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
import msgpack
import pytest
import zmq

from mercury.common.exceptions import MercuryClientException
from mercury.common import transport
from .base import MercuryCommonUnitTest


@mock.patch.object(transport, 'get_ctx_and_connect_req_socket')
def test_full_req_transceiver(mock_get_ctx):
    ctx = mock.Mock(spec_set=zmq.Context)
    socket = ctx.socket.return_value
    mock_get_ctx.return_value = ctx, socket
    socket.recv.return_value = msgpack.packb('response')

    unpacked = transport.full_req_transceiver('localhost', 'data')

    assert unpacked == 'response'
    socket.send_multipart.assert_called_once_with([msgpack.packb('data')])
    socket.recv.assert_called_once()
    socket.close.assert_called_once()
    ctx.term.assert_called_once()


def test_get_ctx_and_connect_req_socket():
    with mock.patch('zmq.Context.socket') as mock_socket:
        ctx, socket = transport.get_ctx_and_connect_req_socket('zmq_url')

    assert isinstance(ctx, zmq.Context)
    mock_socket.assert_called_once()


def test_parse_multipart_message():
    """Test parse_multipart_message()"""
    message = ['addr', '', 'msg']
    parsed_message = transport.parse_multipart_message(message)
    assert parsed_message == {'address': ['addr', ''], 'message': 'msg'}


def test_parse_multipart_message_error():
    """Test parse_multipart_message() with a 2-parts message."""
    message = ['addr', '']
    parsed_message = transport.parse_multipart_message(message)
    assert parsed_message == {}


def test_serialize_addresses():
    """Test serialize_addresses()"""
    multipart_address = ['addr', 'addr2', '']
    addresses = transport.serialize_addresses(multipart_address)
    assert addresses == ['addr', 'addr2']


class SimpleRouterReqServiceUnitTest(MercuryCommonUnitTest):
    """Tests for mercury.common.transport.SimpleRouterReqService."""
    def setUp(self):
        super(SimpleRouterReqServiceUnitTest, self).setUp()
        self.patch_context = mock.patch('mercury.common.transport.zmq.Context')
        self.mock_context = mock.Mock(spec=zmq.Context)
        mock_context_callable = self.patch_context.start()
        mock_context_callable.return_value = self.mock_context
        self.req_service = transport.SimpleRouterReqService('localhost')

    def tearDown(self):
        super(SimpleRouterReqServiceUnitTest, self).tearDown()
        self.patch_context.stop()

    def test_receive(self):
        """Test receive()"""
        payload = {'endpoint': 'insert_one', 'args': ['fake_info']}
        multipart_message = ['addr', '', msgpack.packb(payload)]
        self.req_service.socket.recv_multipart.return_value = multipart_message

        received_msg = self.req_service.receive()

        expected_msg = (['addr', ''],
                        {'endpoint': 'insert_one', 'args': ['fake_info']})
        assert received_msg == expected_msg

    def test_receive_no_message(self):
        """Test receive() fails when received message is empty."""
        self.req_service.socket.recv_multipart.return_value = []
        self.assertRaises(MercuryClientException, self.req_service.receive)

    @mock.patch('msgpack.unpackb')
    def test_receive_wrong_type(self, mock_unpack):
        """Test receive() fails when received message cannot be unpacked"""
        payload = {'endpoint': 'insert_one', 'args': ['fake_info']}
        multipart_message = ['addr', '', payload]
        self.req_service.socket.recv_multipart.return_value = multipart_message

        mock_unpack.side_effect = [TypeError]

        with mock.patch.object(self.req_service, 'send_error'):
            self.assertRaises(MercuryClientException,
                              self.req_service.receive)

    @mock.patch('msgpack.unpackb')
    def test_receive_unpack_error(self, mock_unpack):
        """Test receive() fails when unpacking fails"""
        payload = {'endpoint': 'insert_one', 'args': ['fake_info']}
        multipart_message = ['addr', '', payload]
        self.req_service.socket.recv_multipart.return_value = multipart_message

        mock_unpack.side_effect = [msgpack.UnpackException]

        with mock.patch.object(self.req_service, 'send_error') as mock_error:
            self.assertRaises(MercuryClientException,
                              self.req_service.receive)

    def test_send_error(self):
        address = ['addr', '']
        message = {'endpoint': 'insert_one'}

        with mock.patch.object(self.req_service, 'send') as mock_send:
            self.req_service.send_error(address, message)

        mock_send.assert_called_once_with(address,
                                          {'error': True, 'message': message})

    def test_send(self):
        address = ['addr', '']
        data = {'error': True, 'message': 'fake_message'}

        expected_arg = ['addr', '', msgpack.packb(data)]

        self.req_service.send(address, data)
        self.req_service.socket.send_multipart.assert_called_once_with(
            expected_arg)

    def test_get_key(self):
        """Test get_key()"""
        data = {'key': 'value'}

        returned_value = self.req_service.get_key('key', data)
        assert returned_value == 'value'

    def test_get_key_error(self):
        """Test get_key() when key is missing from the dict"""
        data = {'key': 'value'}

        with pytest.raises(MercuryClientException):
            self.req_service.get_key('wrong_key', data)

    def test_validate_required_error(self):
        """Test validate_required() when a key is missing."""
        required = ['key1', 'key2']
        data = {'key1': 'val1'}

        with pytest.raises(MercuryClientException) as exc_info:
            self.req_service.validate_required(required, data)
            if hasattr(exc_info.value, 'message'):
                assert exc_info.value.message == (
                    "Message is missing required data: ['key2']")
            else:
                assert exc_info.value.msg == \
                       "Message is missing required data: ['key2']"



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""\
Functions for grabbing a MercuryID

MercuryID or mercury_id meta and hash value


[meta] - 00 = hash generated by interface mac addresses
         01 = hash generated by asset tag and serial number
"""
import hashlib
import logging

from mercury.common.exceptions import MercuryIdException

LOG = logging.getLogger(__name__)

META_TYPE_MAC = '00'
META_TYPE_PRODUCT_UUID = '01'
META_TYPE_CHASSIS_ASSET_SERIAL = '02'
META_TYPE_BOARD_ASSET_SERIAL = '03'


def _build_hash(target, meta_type):
    """Builds hash from target data and meta_type"""
    digest = hashlib.sha1(target.encode('ascii')).hexdigest()
    return meta_type + digest


def _get_embedded(inspected_interfaces):
    """Gets embedded interfaces from inspected interfaces."""
    embedded_interfaces = []
    for interface in inspected_interfaces:
        _biosdevname = interface['predictable_names'].get('biosdevname', '')
        if _biosdevname:
            if 'em' in _biosdevname:
                embedded_interfaces.append(interface)
    return embedded_interfaces


DMI_DISQUALIFIED_STRING = 'To Be Filled By O.E.M.'


def _dmi_methods(dmi):
    """Builds a mercury ID for DMI information if possible.

    :param dmi: A dictionary of DMI information from a system.
    :returns: A string if the right DMI is present, None otherwise.
    """
    product_uuid = dmi.get('product_uuid')
    chassis_asset_tag = dmi.get('chassis_asset_tag')
    chassis_serial = dmi.get('chassis_serial')
    board_asset_tag = dmi.get('board_asset_tag')
    board_serial = dmi.get('board_serial')

    if product_uuid:
        LOG.debug('Generating mercury ID using product_uuid: %s' % product_uuid)
        return _build_hash(product_uuid, META_TYPE_PRODUCT_UUID)

    if DMI_DISQUALIFIED_STRING in [chassis_asset_tag, chassis_serial,
                                   board_asset_tag, board_serial]:
        LOG.debug('Junk in DMI tables: \'%s\'' % DMI_DISQUALIFIED_STRING)
        return

    if chassis_asset_tag and chassis_serial:
        LOG.debug('Generating mercury ID using chassis asset information: tag=%s, asset=%s' % (
            chassis_asset_tag, chassis_serial))
        return _build_hash(chassis_asset_tag + chassis_serial, META_TYPE_CHASSIS_ASSET_SERIAL)

    if board_asset_tag and board_serial:
        LOG.debug('Generating mercury ID using board asset information: tag=%s, asset=%s' % (
                  board_asset_tag, board_serial))
        return _build_hash(board_asset_tag + board_serial, META_TYPE_BOARD_ASSET_SERIAL)


def generate_mercury_id(inspected_dmi, inspected_interfaces):
    """Generates a mercury ID based on gathered system information.

    :param inspected_dmi: A dictionary containing DMI information about
        the system in question.
    :param inspected_interfaces: A dictionary containing information about
        interfaces present in the system.
    :returns: A string representing the mercury ID.
    :raises MercuryIdException: If not enough information is present to
        generate an ID.
    """
    mercury_id = _dmi_methods(inspected_dmi)
    if mercury_id:
        return mercury_id
    else:
        meta_type = META_TYPE_MAC
        embedded = _get_embedded(inspected_interfaces)
        if embedded:
            LOG.debug('Generating mercury ID using embedded interfaces ')
            inspected_interfaces = embedded
        else:
            LOG.debug('Generating mercury ID using all interfaces')

        target = ''
        for interface in inspected_interfaces:
            address = interface.get('address')  # mac address
            if address:
                target += address

    if not target:
        raise MercuryIdException('Could not generate MercuryId')

    return _build_hash(target, meta_type)



from mercury.common.configuration import MercuryConfiguration

from mercury.backend.configuration import BACKEND_CONFIG_FILE


def parse_options():
    configuration = MercuryConfiguration(
        'mercury-queue-backend',
        BACKEND_CONFIG_FILE,
        description='Mercury backend queue service'
    )

    configuration.add_option('backend.queue_service.bind_address',
                             default='tcp://127.0.0.1:9007',
                             help_string='Queue backend bind address')

    configuration.add_option('backend.redis.host',
                             default='localhost',
                             help_string='Redis server address')

    configuration.add_option('backend.redis.port',
                             default='6379',
                             special_type=int,
                             help_string='Redis server port')

    configuration.add_option('backend.redis.queue',
                             default='rpc_task_queue',
                             help_string='The queue to use for RPC tasks')

    configuration.add_option('backend.inventory_router',
                             default='tcp://localhost:9000',
                             help_string='Inventory router url')

    return configuration.scan_options()



# Copyright 2018 Ruben Quinones (ruben.quinones@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from flask import Flask

from mercury.boot.configuration import get_boot_configuration
from mercury.boot.urls import boot_urls

app = Flask(__name__)

# Add url rules
for url, view_func in boot_urls:
    app.add_url_rule(url, view_func=view_func, strict_slashes=False)

if __name__ == '__main__':
    configuration = get_boot_configuration()
    app.run(host=configuration.host, port=configuration.port, debug=True)



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from mercury.common.task_managers.base import task
from tests.common.unit.base import MercuryCommonUnitTest


class TaskTest(MercuryCommonUnitTest):
    """Tests for the mercury.common.task_managers.base.task"""
    def setUp(self):
        """Create a fake Task object"""
        self.task = task.Task()
        pass

    @mock.patch.object(task.Task, 'do')
    def test_execute(self, mock_do):
        """Test execute() calls do() and reset the task to None"""
        self.task.task = 'fake_task'

        self.task.execute()
        mock_do.assert_called_once()
        self.assertIsNone(self.task.task)

    def test_execute_no_task(self):
        """Test execute() fails when no task is defined"""
        self.assertRaises(Exception, self.task.execute)



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
import pytest
import six
import subprocess

from mercury.common.helpers import cli
from tests.common.unit.base import MercuryCommonUnitTest


class CLIResultUnitTest(MercuryCommonUnitTest):
    """Unit tests for mercury.common.helpers.cli.CLIResult"""
    def test_setter(self):
        """Test setters for stdout and stderr."""
        result = cli.CLIResult('', '', 0)
        result.stdout = 'out'
        result.stderr = 'err'

        assert repr(result) == 'out'
        assert result.stderr == 'err'

    def test_decode(self):
        """Test arguments are decoded for new CLIResult instances."""
        output = 'out'
        if isinstance(output, bytes):
            output_str = output.decode(cli.CLIResult.ENCODING)
            output_bytes = output
        else:
            output_str = output
            output_bytes = output.encode(cli.CLIResult.ENCODING)

        result_str = cli.CLIResult(output_str, 'err', 0)
        result_bytes = cli.CLIResult(output_bytes, 'err', 0)

        assert isinstance(result_str.stdout, six.string_types)
        assert isinstance(result_bytes.stdout, six.string_types)


class CliRunUnitTest(MercuryCommonUnitTest):
    """Unit tests for mercury.common.helpers.cli.run()"""
    @mock.patch('mercury.common.helpers.cli.os.environ.copy')
    @mock.patch('subprocess.Popen')
    def test_run(self, mock_subprocess_popen, mock_os_environ):
        """Test run()"""
        mock_subprocess_popen.return_value.returncode = 0
        mock_subprocess_popen.return_value.communicate.return_value = ('out',
                                                                       'err')
        mock_os_environ.return_value = {'key': 'val'}

        cli_result = cli.run('ls')

        mock_subprocess_popen.assert_called_once_with(['ls'],
                                                      stdin=None,
                                                      stdout=subprocess.PIPE,
                                                      stderr=subprocess.PIPE,
                                                      bufsize=1048567,
                                                      env={'key': 'val'})
        mock_subprocess_popen.return_value.communicate.assert_called_once_with(
            input=None)
        assert cli_result.stdout == 'out'
        assert cli_result.stderr == 'err'
        assert cli_result.returncode == 0

    @mock.patch('mercury.common.helpers.cli.os.environ.copy')
    @mock.patch('subprocess.Popen')
    def test_run_with_input(self, mock_subprocess_popen, mock_os_environ):
        """Test run() with data passed into stdin."""
        mock_subprocess_popen.return_value.returncode = 0
        mock_subprocess_popen.return_value.communicate.return_value = ('foo',
                                                                       '')
        mock_os_environ.return_value = {'key': 'val'}

        cli_result = cli.run('python', _input='print "foo"')

        mock_subprocess_popen.assert_called_once_with(['python'],
                                                      stdin=subprocess.PIPE,
                                                      stdout=subprocess.PIPE,
                                                      stderr=subprocess.PIPE,
                                                      bufsize=1048567,
                                                      env={'key': 'val'})
        mock_subprocess_popen.return_value.communicate.assert_called_once_with(
            input='print "foo"')
        assert cli_result.stdout == 'foo'
        assert cli_result.stderr == ''
        assert cli_result.returncode == 0

    @mock.patch('subprocess.Popen')
    def test_run_dry_run(self, mock_subprocess_popen):
        """Test run() in dry_run mode."""
        cli_result = cli.run('ls', dry_run=True)

        mock_subprocess_popen.assert_not_called()
        mock_subprocess_popen.return_value.communicate.assert_not_called()
        assert cli_result.stdout == ''
        assert cli_result.stderr == ''
        assert cli_result.returncode == 0

    @mock.patch('subprocess.Popen')
    def test_run_popen_exception(self, mock_subprocess_popen):
        """Test run() with subprocess.Popen failing."""
        mock_subprocess_popen.side_effect = [OSError, ValueError]

        with pytest.raises(cli.CLIException):
            cli.run('ls', raise_exception=True)

        cli_result = cli.run('ls', raise_exception=False)
        assert cli_result.stdout == ''
        assert cli_result.stderr == "Failed while executing 'ls': "
        assert cli_result.returncode == 1

        assert mock_subprocess_popen.call_count == 2
        mock_subprocess_popen.return_value.communicate.assert_not_called()

    @mock.patch('subprocess.Popen')
    def test_run_error(self, mock_subprocess_popen):
        """Test run() when the command returns an error."""
        mock_subprocess_popen.return_value.returncode = 1
        mock_subprocess_popen.return_value.communicate.return_value = ('',
                                                                       'err')

        with pytest.raises(cli.CLIException):
            cli.run('ls', raise_exception=True)

        cli_result = cli.run('ls', raise_exception=False)
        assert cli_result.stdout == ''
        assert cli_result.stderr == 'err'
        assert cli_result.returncode == 1

        assert mock_subprocess_popen.call_count == 2
        assert mock_subprocess_popen.return_value.communicate.call_count == 2


@mock.patch('mercury.common.helpers.cli.os.path.exists')
def test_find_in_path(mock_os_path_exists):
    """Test find_in_path()."""
    # Test with absolute path.
    mock_os_path_exists.return_value = True
    path = cli.find_in_path('/does/not/exist')
    assert path == '/does/not/exist'

    # Test with relative path.
    with mock.patch.dict('os.environ', {'PATH': '/:/tmp'}):
        path = cli.find_in_path('foo')
        assert path == '/foo'

    # Test with non-existing file.
    mock_os_path_exists.return_value = False
    path = cli.find_in_path('/tmp/foo')
    assert path is None



"""
Commonly used exceptions used in tests.
"""


class BaseBehaveException(Exception):
    message = "Not Set"

    def __init__(self, message=None):
        super(BaseBehaveException, self).__init__()
        self.message = message or self.message

    def __str__(self):
        return repr(self.message)


class TimeoutException(BaseBehaveException):
    message = "Request timed out"



import mock
import pytest
import unittest

from mercury.common.exceptions import MercuryClientException
from mercury.log_service import service as log_server

mock_make_agent_log_service = mock.MagicMock()


def mock_init(self, bind_address):
    self.bind_address = bind_address
    self.context = mock.MagicMock()
    self.socket = mock.MagicMock()
    self.bound = False

class AgentLogServiceTestCase(unittest.TestCase):
    def setUp(self):
        with mock.patch(
                'mercury.log_service.service.SimpleRouterReqService.__init__',
                new=mock_init):
            log_collection = mock.Mock()
            self.agent_log_service = log_server.AgentLogService('invalid_bind_url',
                                                                log_collection)

    def test_process(self):
        with pytest.raises(MercuryClientException):
            self.agent_log_service.process({})

        message = {
            "name": "mercury.agent.agent",
            "msg": "Injection completed",
            "args": [],
            "levelname": "INFO",
            "levelno": 20,
            "pathname": "/src/mercury/agent/mercury/agent/agent.py",
            "filename": "agent.py",
            "module": "agent",
            "exc_info": None,
            "exc_text": None,
            "stack_info": None,
            "lineno": 89,
            "funcName": "run",
            "created": 1510076870.9318643,
            "msecs": 931.8642616271973,
            "relativeCreated": 405.81798553466797,
            "thread": 139861930071808,
            "threadName": "MainThread",
            "processName": "MainProcess",
            "process": 71,
            "message": "Injection completed",
            "asctime": "2017-11-07 17:47:50,931",
            "mercury_id": "0112426690c16b0fd4586e91731a0cc4f3f918c5af"
        }
        self.agent_log_service.log_collection.insert = mock.MagicMock()
        assert self.agent_log_service.process(message) == {'message': 'ok', 'error': False}
        self.agent_log_service.log_collection.insert.assert_called_once_with(message)




import logging

log = logging.getLogger(__name__)


class Task(object):
    def __init__(self):
        self.task = None

    def fetch(self):
        raise NotImplementedError

    def do(self):
        raise NotImplementedError

    def execute(self):
        if not self.task:
            raise Exception('Task has not been populated')
        result = self.do()
        self.task = None
        return result



class ConfigVars(object):
    """Configuration environment variables"""

    # behavior
    VERBOSE = 'VERBOSE'

    # Mercury API URL
    MERCURY_API_ENDPOINT = 'MERCURY_API_ENDPOINT'
    # Identity
    INTERNAL_IDENTITY_URL = 'INTERNAL_IDENTITY_URL'
    INTERNAL_IDENTITY_USERNAME = 'INTERNAL_IDENTITY_USERNAME'
    INTERNAL_IDENTITY_PASSWORD = 'INTERNAL_IDENTITY_PASSWORD'
    DOMAIN = 'DOMAIN'
    DOMAIN_NAME = 'DOMAIN_NAME'
    # Test Data
    JSON_API_DATA_LOCATION = 'JSON_API_DATA_LOCATION'

class Sources(object):
    """Sources to get the configuration variables from"""

    ENVIRON = 'environ'



from mercury.common.clients.router_req_client import RouterReqClient
from mercury.common.asyncio.clients.async_router_req_client import \
    AsyncRouterReqClient


class RPCClient(RouterReqClient):
    service_name = 'RPC Service'

    def update_task(self, update_data):
        """

        :param update_data:
        :return:
        """
        return self.transceiver({
            'endpoint': 'update_task',
            'args': [update_data]
        })

    def complete_task(self, result_data):
        """
        :param result_data:
        :return:
        """
        return self.transceiver({
            'endpoint': 'complete_task',
            'args': [result_data]
        })


class AsyncRPCClient(AsyncRouterReqClient):
    service_name = 'RPC client'

    async def update_task(self, update_data):
        """

        :param update_data:
        :return:
        """
        return await self.transceiver({
            'endpoint': 'update_task',
            'args': [update_data]
        })

    async def complete_task(self, result_data):
        """
        :param result_data:
        :return:
        """
        return await self.transceiver({
            'endpoint': 'complete_task',
            'args': [result_data]
        })



# Copyright 2018 Ruben Quinones (ruben.quinones@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mercury.boot.views import BootView, DiscoverView

boot_view = BootView.as_view('boot')
discover_view = DiscoverView.as_view('discover')

boot_urls = [
    # Boot url rules
    ('/boot', boot_view),
    # Discover rules
    ('/boot/discover/<mac_address>', discover_view),
]



from mercury.common.configuration import MercuryConfiguration

DEFAULT_CONFIG_FILE = 'mercury-inventory.yaml'


def parse_options():
    configuration = MercuryConfiguration(
        'mercury-inventory',
        DEFAULT_CONFIG_FILE,
        description='The mercury inventory service'
    )

    configuration.add_option('inventory.bind_address',
                             default='tcp://127.0.0.1:9000',
                             help_string='Mercury Inventory bind address'
                             )

    configuration.add_option('asyncio_debug',
                             '--asyncio-debug',
                             'ASYNCIO_DEBUG',
                             'logging.debug_asyncio',
                             default=False,
                             special_type=bool,
                             help_string='Enable asyncio debugging'
                             )

    configuration.add_option('inventory.database.name',
                             default='test',
                             help_string='The inventory database')

    configuration.add_option('inventory.database.collection',
                             default='inventory',
                             help_string='The collection for inventory '
                                         'documents')

    configuration.add_option('inventory.database.servers',
                             default='localhost:27017',
                             special_type=list,
                             help_string='Server or coma separated list of '
                                         'servers to connect to')

    configuration.add_option('inventory.database.replica_name',
                             help_string='An optional replica name')

    configuration.add_option('inventory.database.username',
                             help_string='The database user')

    configuration.add_option('inventory.database.password',
                             help_string='The database user\'s password')

    return configuration.scan_options()



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mercury.common.exceptions import MercuryClientException
from mercury.common.asyncio.clients.async_router_req_client import \
    AsyncRouterReqClient


class InventoryClient(AsyncRouterReqClient):
    """Client to interact with inventory."""

    service_name = 'Inventory'

    async def insert_one(self, device_info):
        """Insert a new device in inventory.

        :param device_info: A dict containing the new device info.
        :returns: The 'response' field of the transceiver's reply.
        :raises: MercuryClientException if 'mercury_id' is missing.
        """
        mercury_id = device_info.get('mercury_id')
        if not mercury_id:
            raise MercuryClientException('device_info is missing mercury_id')

        payload = {
            'endpoint': 'insert_one',
            'args': [device_info]
        }
        return await self.transceiver(payload)

    async def update_one(self, mercury_id, update_data):
        """Update a device in inventory.

        :param mercury_id: The ID of the device to update.
        :param update_data: A dict containing the data to update.
        :returns: The 'response' field of the transceiver's reply.
        """
        payload = {
            'endpoint': 'update_one',
            'args': [mercury_id],
            'kwargs': {'update_data': update_data}
        }

        return await self.transceiver(payload)

    async def get_one(self, mercury_id, projection=None):
        """Get a device from inventory.

        :param mercury_id: The ID of the device.
        :param projection: A dict specifying which fields should be included
            in the results.
        :returns: The 'response' field of the transceiver's reply.
        """
        payload = {
            'endpoint': 'get_one',
            'args': [mercury_id],
            'kwargs': {
                'projection': projection
            }
        }
        return await self.transceiver(payload)

    async def query(self, query_data, projection=None, limit=0, sort='_id',
                    sort_direction=1, offset_id=None):
        """Query inventory for devices matching query_data.

        :param query_data: A dict to filter the results.
        :param projection: A dict specifying which fields should be included
            in the results.
        :param sort:
        :param limit: The maximum number of results to return.
        :param sort_direction: The sort direction.
        :param offset_id:
        :returns: The 'response' field of the transceiver's reply.
        """
        payload = {
            'endpoint': 'query',
            'args': [query_data],
            'kwargs': {
                'projection': projection,
                'limit': limit,
                'sort_direction': sort_direction
            }
        }
        return await self.transceiver(payload)

    async def delete(self, mercury_id):
        """Delete a device from inventory.

        :param mercury_id: The ID of the device.
        :returns: The 'response' field of the transceiver's reply.
        """
        payload = {
            'endpoint': 'delete',
            'args': [mercury_id]
        }
        return await self.transceiver(payload)

    async def count(self, query_data):
        """Count how many devices match query_data.

        :param query_data: A dict to filter the results.
        :returns: The 'response' field of the transceiver's reply.
        """
        payload = {
            'endpoint': 'count',
            'args': [query_data]
        }
        return await self.transceiver(payload)



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import mock
import pytest

from mercury.rpc.controller import RPCController


@pytest.fixture()
def frontend_controller(async_mongodb):
    jobs_collection = async_mongodb.rpc_jobs
    tasks_collection = async_mongodb.rpc_tasks
    inventory_client = mock.Mock()
    controller = RPCController(inventory_client,
                               jobs_collection,
                               tasks_collection)
    return controller


class TestFrontendController(object):

    @pytest.mark.asyncio
    async def test_get_job(self, frontend_controller):
        job = await frontend_controller.get_job('job-1')
        assert job.get('job_id') == 'job-1'

    @pytest.mark.asyncio
    async def test_get_job_none(self, frontend_controller):
        job = await frontend_controller.get_job('job-x')
        assert job == None



import logging
import signal
import threading
import time

from .worker import Worker


log = logging.getLogger(__name__)


class Manager(object):
    """Class managing workers"""
    def __init__(self,
                 task_handler,
                 number_of_workers=10,
                 maximum_requests_per_thread=5000,
                 maximum_age=3600,
                 handler_args=None,
                 handler_kwargs=None):
        """Create a new Manager.

        :param task_handler: The Task object that will execute tasks.
        :param number_of_workers: Maximum number of workers for this manager.
        :param maximum_requests_per_thread: Maximum number of tasks per worker.
        :param maximum_age: Maximum age of workers.
        """
        self.number_of_workers = number_of_workers
        self.max_requests = maximum_requests_per_thread
        self.max_age = maximum_age
        self.task_handler = task_handler
        self.handler_args = handler_args or ()
        self.handler_kwargs = handler_kwargs or {}
        self.workers = []

    def spawn(self):
        """Add and start a new worker to handle tasks."""
        worker_dict = {}
        w = Worker(self.task_handler, self.max_requests, self.max_age,
                   self.handler_args, self.handler_kwargs)
        worker_dict['worker_class'] = w
        t = threading.Thread(target=w.start)
        t.start()
        log.info('Spawned thread: %s' % t.ident)
        worker_dict['thread'] = t
        self.workers.append(worker_dict)

    def kill_all(self):
        """Kill all active workers."""
        for worker in self.active_workers:
            log.info('Sending kill signal to worker %s'
                     % worker['thread'].ident)
            worker['worker_class'].kill_signal = True

    @property
    def active_workers(self):
        """Find all active workers and update current list."""
        active = []
        for idx in range(len(self.workers)):
            if self.workers[idx]['thread'].is_alive():
                active.append(self.workers[idx])
        # An optimization that usually should be present in a property
        # "I'm sure you can find a better way to implement this" - Hussam
        self.workers = active
        return active

    def spawn_threads(self):
        """Create new workers until the maximum number is reached."""
        count = len(self.active_workers)
        while count < self.number_of_workers:
            self.spawn()
            count += 1

    def manage(self):
        """Maintain the maximum number of workers running."""
        # noinspection PyUnusedLocal
        def term_handler(*args, **kwargs):
            log.info('Caught TERM signal, sending kill all')
            self.kill_all()

        signal.signal(signal.SIGTERM, term_handler)

        try:
            while True:
                self.spawn_threads()
                time.sleep(.01)
        except KeyboardInterrupt:
            term_handler()



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from mercury.common.task_managers.base import task
from mercury.common.task_managers.base import worker
from tests.common.unit.base import MercuryCommonUnitTest


class WorkerTest(MercuryCommonUnitTest):
    """Tests for the mercury.common.task_managers.base.worker"""
    @mock.patch.object(task, 'Task')
    def setUp(self, mock_task):
        """Create a fake worker object"""
        mock_task.create.return_value = None

        task_handler = task.Task()
        max_requests = 1
        max_age = 3600
        self.fake_worker = worker.Worker(task_handler, max_requests, max_age,
                                         (), {})

    def test_start(self):
        """Test start() executes a new task"""
        self.fake_worker.start()
        self.fake_worker.task.fetch.assert_called_once()
        self.fake_worker.task.execute.assert_called_once()
        self.assertEqual(1, self.fake_worker.handled_tasks)

    def test_start_kill_signal(self):
        """Test start() doesn't execute task if kill_signal is True"""
        self.fake_worker.kill_signal = True
        self.fake_worker.start()
        self.fake_worker.task.fetch.assert_not_called()
        self.fake_worker.task.execute.assert_not_called()
        self.assertEqual(0, self.fake_worker.handled_tasks)

    def test_start_too_many_requests(self):
        """Test start() doesn't execute more tasks than maximum allowed"""
        self.fake_worker.handled_tasks = 1
        self.fake_worker.start()
        self.fake_worker.task.fetch.assert_not_called()
        self.fake_worker.task.execute.assert_not_called()
        self.assertEqual(1, self.fake_worker.handled_tasks)

    def test_start_no_more_task(self):
        """Test start() continue fetching tasks if none found at first"""
        self.fake_worker.task.fetch.side_effect = [None, 'fake_task']
        self.fake_worker.start()
        self.assertEqual(2, self.fake_worker.task.fetch.call_count)
        self.fake_worker.task.execute.assert_called_once()
        self.assertEqual(1, self.fake_worker.handled_tasks)



import logging

import msgpack
import zmq
import zmq.asyncio

from mercury.common.exceptions import (
    parse_exception,
    fancy_traceback_short,
    MercuryTransportError
 )


log = logging.getLogger(__name__)


class AsyncRouterReqClient(object):
    service_name = 'Generic Service'

    def __init__(self, zmq_url, linger=-1, response_timeout=-1, rcv_retry=0,
                 raise_on_timeout=False, keep_alive_interval=0):
        """

        :param zmq_url:
        :param linger:
        :param response_timeout:
        :param rcv_retry: The number of times to retry
        :param keep_alive_interval: The interval to send keep_alive messages in
        seconds
        """
        self.zmq_url = zmq_url
        self.linger = linger
        self.response_timeout = response_timeout
        self.rcv_retry = rcv_retry
        self.raise_on_timeout = raise_on_timeout

        self.context = zmq.asyncio.Context()

        self.socket = None
        self.refresh_socket()

    def refresh_socket(self):
        """
        Creates a new socket if socket is None or the socket is closed
        :return:
        """
        if not self.socket or self.socket.closed:
            self.socket = self.context.socket(zmq.REQ)

            # < -1 is not supported, so assume anything less than 0 should be
            # -1
            if self.response_timeout < 0:
                timeout = -1
            else:
                # positive timeouts should be expressed in milliseconds
                timeout = self.response_timeout * 1000

            self.socket.setsockopt(zmq.LINGER, self.linger)
            self.socket.setsockopt(zmq.RCVTIMEO, timeout)

            # Keep Alive

            self.socket.setsockopt(zmq.TCP_KEEPALIVE, 1)
            self.socket.setsockopt(zmq.TCP_KEEPALIVE_IDLE, 120)
            self.socket.setsockopt(zmq.TCP_KEEPALIVE_CNT, 3)
            self.socket.setsockopt(zmq.TCP_KEEPALIVE_INTVL, 10)

            self.socket.connect(self.zmq_url)

    def safe_send(self, data):
        """
        Ensures that the socket is alive by calling refresh_socket
        :return:
        """
        # Refresh the socket if necessary
        self.refresh_socket()

        # Pack data
        packed = msgpack.packb(data)

        # Send does not block
        self.socket.send_multipart([packed])

    async def safe_receive(self):
        """
        Receive server reply.
        :return: The unpacked message
        """
        retry_count = self.rcv_retry and self.rcv_retry + 1 or 1

        while retry_count:
            # noinspection PyBroadException
            try:
                return msgpack.unpackb(
                    await self.socket.recv(), encoding='utf-8')
            except zmq.Again:
                retry_count -= 1
                log.error(f'[{self.service_name}] Receive timeout'
                          f' Retries remaining: {retry_count}')
            # TODO: explicitly catch things that zmq or msgpack might throw
            # Catchall
            except Exception:
                log.error(f'An unhandled exception occurred while receiving'
                          f' <{fancy_traceback_short(parse_exception())}>')
                break

        # Errors should be actionable
        error_message = f'[{self.service_name}] Failure receiving server ' \
                        f'reply. Closing socket to reset state.'
        log.error(error_message)
        self.close()

        if self.raise_on_timeout:
            raise MercuryTransportError(error_message)

        return {'error': True, 'message': error_message}

    async def transceiver(self, payload):
        """
        Convenience method
        :param payload:
        :return:
        """
        self.safe_send(payload)
        return await self.safe_receive()

    def close(self):
        """ close the socket """
        self.socket.close()



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import asyncio
import logging

import zmq
import zmq.asyncio

from mercury.common.asyncio.dispatcher import AsyncDispatcher
from mercury.common.asyncio.transport import TrivialAsyncRouterReqService
from mercury.common.asyncio.clients.inventory import \
    InventoryClient as AsyncInventoryClient
from mercury.common.clients.inventory import InventoryClient
from mercury.backend.active_asyncio import add_active_record, ping_loop, \
    stop_ping
from mercury.backend.controller import BackendController
from mercury.backend.options import parse_options
from mercury.backend.rpc_client import AsyncRPCClient

log = logging.getLogger(__name__)


class BackEndService(TrivialAsyncRouterReqService):
    def __init__(self,
                 bind_address,
                 inventory_client,
                 rpc_client,
                 name,
                 datacenter,
                 vip,
                 port):
        super(BackEndService, self).__init__(bind_address)

        self.inventory_client = inventory_client
        self.rpc_client = rpc_client
        self.server_info = {
            'name': name,
            'datacenter': datacenter,
            'address': vip,
            'port': port
        }
        self.controller = BackendController(self.server_info,
                                            self.inventory_client,
                                            self.rpc_client)

        self.dispatcher = AsyncDispatcher(self.controller)

    async def process(self, message):
        """ Process the message via dispatcher """
        return await self.dispatcher.dispatch(message)


def reacquire(inventory_url, backend_name):
    """

    :param inventory_url:
    :param backend_name:
    :return:
    """
    # Onetime use synchronous client
    log.info('Attempting to reacquire active agents')
    log.debug('Inventory Router: {}'.format(inventory_url))

    inventory_client = InventoryClient(inventory_url,
                                       # TODO: Add these to configuration
                                       response_timeout=60,
                                       rcv_retry=10)

    existing_documents = inventory_client.query({'active': {'$ne': None},
                                                 'origin.name': backend_name},
                                                projection={'mercury_id': 1,
                                                            'active': 1})

    if existing_documents.get('error'):  # Transport Error
        log.error('[BACKEND CRITICAL] '
                  'Error communicating with inventory service, could not '
                  'reacquire: <{}>'.format(existing_documents.get('message')))
        # Return without reacquiring any nodes. Once communication is
        # reestablished, agents will begin to re-register
        return

    for doc in existing_documents['message']['items']:
        if not BackendController.validate_agent_info(doc['active']):
            log.error('Found junk in document {} expunging'.format(
                doc['mercury_id']))
            inventory_client.update_one(doc['mercury_id'], {'active': None})

        log.info('Attempting to reacquire %s : %s' % (
            doc['mercury_id'], doc['active']['rpc_address']))
        add_active_record(doc)

    log.info('Reacquire operation complete')
    inventory_client.close()


def configure_logging(config):
    """ Configure logging for application
    :param config: A namespace provided from MercuryConfiguration.parse_args
    """
    logging.basicConfig(level=logging.getLevelName(config.logging.level),
                        format=config.logging.format)

    if config.subtask_debug:
        logging.getLogger('mercury.rpc.ping').setLevel(logging.DEBUG)
        logging.getLogger('mercury.rpc.ping2').setLevel(logging.DEBUG)
        logging.getLogger('mercury.rpc.jobs.monitor').setLevel(logging.DEBUG)

    if config.asyncio_debug:
        logging.getLogger('mercury.rpc.active_asyncio').setLevel(logging.DEBUG)


def main():
    """ Entry point """

    config = parse_options()

    configure_logging(config)

    # Create the event loop
    loop = zmq.asyncio.ZMQEventLoop()

    # If config.asyncio_debug == True, enable debug
    loop.set_debug(config.asyncio_debug)

    # Set the zmq event loop as the default event loop
    asyncio.set_event_loop(loop)

    # Create Async Clients
    inventory_client = AsyncInventoryClient(config.backend.inventory_router,
                                            linger=0,
                                            response_timeout=10,
                                            rcv_retry=3)
    rpc_client = AsyncRPCClient(config.backend.rpc_router,
                                linger=0,
                                response_timeout=10,
                                rcv_retry=3)

    # Create a backend instance
    server = BackEndService(config.backend.agent_service.bind_address,
                            inventory_client,
                            rpc_client,
                            config.backend.origin.name,
                            config.backend.origin.datacenter,
                            config.backend.origin.queue_service_vip,
                            config.backend.origin.queue_service_port)

    reacquire(config.backend.inventory_router, config.backend.origin.name)

    # Inject ping loop
    asyncio.ensure_future(ping_loop(
        server.context,
        config.backend.ping.interval,
        config.backend.ping.cycle_time,
        config.backend.ping.initial_timeout,
        config.backend.ping.retries,
        config.backend.ping.backoff,
        loop,
        config.backend.inventory_router),
        loop=loop)

    log.info('Starting Mercury Backend Service')
    try:
        loop.run_until_complete(server.start())
    except KeyboardInterrupt:
        # TODO: Add generic backend TERM handler
        log.info('Sending kill signals')
        stop_ping()
        server.kill()
    finally:
        pending = asyncio.Task.all_tasks(loop=loop)
        log.debug('Waiting on {} pending tasks'.format(len(pending)))
        loop.run_until_complete(asyncio.gather(*pending))
        log.debug('Shutting down event loop')
        loop.close()


if __name__ == '__main__':
    main()



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from setuptools import setup, find_packages


def get_version():
    with open('VERSION') as fp:
        return fp.readline().strip()


setup(
    name='mercury-core',
    version=get_version(),
    packages=find_packages(exclude=['tests']),
    url='http://www.mercurysoft.io',
    license='Apache-2.0',
    author='Jared Rodriguez',
    author_email='jared.rodriguez@rackspace.com',
    description='Mercury core services and libraries',
    install_requires=[
        'PyYAML',
        'msgpack-python',
        'pymongo',
        'pyzmq',
        'redis',
        'motor',
        'python-box',
        'requests',
        'lxml',
        'pystache',
        'flask'
    ],
    entry_points={
       'console_scripts': [
           'mercury-inventory = mercury.inventory.service:main',
           'mercury-rpc = mercury.rpc.service:main',
           'mercury-backend = mercury.backend.service:main',
           'mercury-backend-queue = mercury.backend.queue_service.service:main',
           'mercury-backend-worker = mercury.backend.workers.service:main',
           'mercury-log = mercury.log_service.service:main',
           'mercury-boot = mercury.boot.service:main'
       ]
    }
)




import asyncio
import logging

import msgpack
import zmq
import zmq.asyncio

from mercury.common.exceptions import fancy_traceback_short, \
    MercuryClientException, parse_exception
from mercury.common.transport import parse_multipart_message

log = logging.getLogger(__name__)


class AsyncRouterReqService(object):
    def __init__(self, bind_address, linger=-1, poll_timeout=2, loop=None):
        self.bind_address = bind_address
        self.loop = loop
        self.context = zmq.asyncio.Context()
        self.poll_timeout = poll_timeout
        self.socket = self.context.socket(zmq.ROUTER)
        self.socket.setsockopt(zmq.LINGER, linger)

        self.in_poller = zmq.asyncio.Poller()
        self.in_poller.register(self.socket, zmq.POLLIN)

        log.info('Bound to: ' + self.bind_address)

        self.socket.bind(self.bind_address)

        self._kill = False

    async def receive(self):
        multipart = await self.socket.recv_multipart()
        parsed_message = parse_multipart_message(multipart)

        if not parsed_message:
            log.error('Received junk off the wire')
            raise MercuryClientException('Message is malformed')

        try:
            message = msgpack.unpackb(parsed_message['message'],
                                      encoding='utf-8')
        except TypeError as type_error:
            log.error('Received unpacked, non-string type: %s : %s' % (
                type(parsed_message), type_error))
            await self.send_error(parsed_message['address'],
                                  'Client error, message is not packed')
            raise MercuryClientException('Message is malformed')

        except (
                msgpack.UnpackException,
                msgpack.ExtraData) as msgpack_exception:
            log.error('Received invalid request: %s' % str(
                msgpack_exception))

            await self.send_error(parsed_message['address'],
                                  'Client error, message is malformed')
            raise MercuryClientException('Message is malformed')

        return parsed_message['address'], message

    async def send_error(self, address, message):
        data = {'error': True, 'message': message}
        log.error(message)
        await self.send(address, data)

    async def send(self, address, message):
        await self.socket.send_multipart(address + [msgpack.packb(message)])

    async def process(self, message):
        raise NotImplementedError

    async def message_handler(self, address, message):
        if isinstance(message, dict) and message.get('_protocol_message') == 'keep_alive':
            log.debug('Keep alive received from {}'.format(address))
            response = {'_protocol_message': 'keep_alive_confirmed'}
        else:
            # noinspection PyBroadException
            try:
                response = await self.process(message)
            except MercuryClientException as mce:
                return await self.send_error(
                    address,
                    'Encountered client error: {}'.format(
                        mce))
            except Exception:
                exec_dict = parse_exception()
                log.error('process raised an exception and should not have.')
                log.error(fancy_traceback_short(exec_dict))
                return await self.send_error(address,
                                             'Encountered server error, sorry')

        await self.send(address, response)
        log.debug('Sent {}'.format(address))

    async def start(self):
        while not self._kill:
            if not await self.in_poller.poll(self.poll_timeout * 1000):
                continue
            try:
                address, message = await self.receive()
            except MercuryClientException:
                continue

            asyncio.ensure_future(self.message_handler(address, message))

        log.info('Shutting down')

    def kill(self):
        self._kill = True

    @staticmethod
    def get_key(key, data):
        try:
            return data[key]
        except KeyError:
            raise MercuryClientException(
                '{} is missing from request'.format(key))

    @staticmethod
    def validate_required(required, data):
        missing = []
        for key in required:
            if key not in data:
                missing.append(data)

        if missing:
            raise MercuryClientException(
                'Message is missing required data: {}'.format(missing))

    def cleanup(self):
        self.socket.close(0)
        self.context.destroy()


class TrivialAsyncRouterReqService(AsyncRouterReqService):
    """ Identical to it's parent, except that it sends a reply back to the
    client BEFORE calling self.process. This is helpful for clients that want
    to fire and forget, and do not care about the result of their message, only
    that it was received.
    """
    RESPONSE_OBJECT = {"error": False, "message": "accepted"}

    async def message_handler(self, address, message):
        await self.send(address, self.RESPONSE_OBJECT)

        # noinspection PyBroadException
        try:
            # Await message processing, but drop result
            await self.process(message)
        except Exception:
            exec_dict = parse_exception()
            log.error('process raised an exception and should not have.')
            log.error(fancy_traceback_short(exec_dict))

    async def process(self, message):
        raise NotImplementedError



import copy
import json
import requests
from src.tests.behave.common.config import get_conflagration


class APIClient(object):
    def __init__(self, base_url, request_kwargs=None, client_kwargs=None):

        self.cfg = get_conflagration()
        token = self.get_identity_token()

        timeout = None
        ssl_certificate_verify = False
        verbose = True

        if request_kwargs:
            timeout = request_kwargs.get("timeout", timeout)
            ssl_certificate_verify = request_kwargs.get(
                "ssl_certificate_verify", ssl_certificate_verify
            )
            verbose = request_kwargs.get("verbose", verbose)

        self.base_url = base_url
        self.headers = {
            "Content-Type": "application/json",
            "X-Auth-Token": token,
        }

        if client_kwargs:
            if not client_kwargs["authorized"]:
                self.headers["X-Auth-Token"] = ""

        self.verify = ssl_certificate_verify
        self.verbose = verbose

        self.request_kwargs = dict()
        self.request_kwargs["url"] = self.base_url
        self.request_kwargs["headers"] = self.headers
        self.request_kwargs["verify"] = self.verify

        if timeout:
            self.request_kwargs["timeout"] = timeout

    def get(
        self,
        resource_id=None,
        params=None,
        url=None,
        url_suffix=None,
        headers=None,
    ):
        request_kwargs = copy.deepcopy(self.request_kwargs)
        if url:
            request_kwargs["url"] = url
        if resource_id:
            resource_url = request_kwargs["url"]
            request_kwargs["url"] = "{0}/{1}".format(resource_url, resource_id)
        if url_suffix:
            resource_url = request_kwargs["url"]
            request_kwargs["url"] = "{0}/{1}".format(resource_url, url_suffix)
        if params:
            request_kwargs["params"] = params
        if headers:
            request_kwargs["headers"].update(headers)
        resp = requests.get(**request_kwargs)
        if self.verbose:
            print("{0}GET REQUEST{1}".format("*" * 20, "*" * 24))
            print(request_kwargs)
            print("{0}RESPONSE{1}".format("*" * 20, "*" * 27))
            print(resp.content)
            print("*" * 48)
        return resp

    def post(self, data, url_suffix=None, headers=None, resource_id=None):
        request_kwargs = copy.deepcopy(self.request_kwargs)
        if resource_id:
            resource_url = request_kwargs["url"]
            request_kwargs["url"] = "{0}/{1}".format(resource_url, resource_id)
        if url_suffix:
            resource_url = request_kwargs["url"]
            request_kwargs["url"] = "{0}/{1}".format(resource_url, url_suffix)
        if headers:
            request_kwargs["headers"].update(headers)
        request_kwargs["data"] = data
        resp = requests.post(**request_kwargs)
        if self.verbose:
            print("{0}POST REQUEST{1}".format("*" * 20, "*" * 24))
            print(request_kwargs)
            print("{0}RESPONSE{1}".format("*" * 20, "*" * 27))
            print(resp.content)
            print("*" * 48)
        return resp

    def delete(self, url_suffix=None, headers=None, resource_id=None):
        request_kwargs = copy.deepcopy(self.request_kwargs)
        if resource_id:
            resource_url = request_kwargs["url"]
            request_kwargs["url"] = "{0}/{1}".format(resource_url, resource_id)
        if url_suffix:
            resource_url = request_kwargs["url"]
            request_kwargs["url"] = "{0}/{1}".format(resource_url, url_suffix)
        if headers:
            request_kwargs["headers"].update(headers)
        resp = requests.delete(**request_kwargs)
        if self.verbose:
            print("{0}DELETE REQUEST{1}".format("*" * 20, "*" * 24))
            print(request_kwargs)
            print("{0}RESPONSE{1}".format("*" * 20, "*" * 27))
            print(resp.content)
            print("*" * 48)
        return resp

    def get_identity_token(self):
        internal_idenity_url = self.cfg.INTERNAL_IDENTITY.internal_identity_url
        internal_identity_username = (
            self.cfg.INTERNAL_IDENTITY.internal_identity_username
        )
        internal_identity_password = (
            self.cfg.INTERNAL_IDENTITY.internal_identity_password
        )
        domain = self.cfg.INTERNAL_IDENTITY.domain
        domain_name = self.cfg.INTERNAL_IDENTITY.domain_name
        identity_headers = {"Content-Type": "application/json"}
        identity_data = {
            "auth": {
                "passwordCredentials": {
                    "username": internal_identity_username,
                    "password": internal_identity_password,
                },
                domain: {"name": domain_name},
            }
        }
        token_resp = requests.post(
            internal_idenity_url,
            headers=identity_headers,
            data=json.dumps(identity_data),
        )
        token = token_resp.json()["access"]["token"]["id"]
        return token



import logging

from mercury.common.configuration import MercuryConfiguration
from mercury.common.task_managers.base.manager import Manager
from mercury.common.task_managers.redis.task import RedisTask
from mercury.common.clients.router_req_client import RouterReqClient

from mercury.backend.rpc_client import RPCClient
from mercury.backend.configuration import (
    add_common_options, BACKEND_CONFIG_FILE
)

log = logging.getLogger(__name__)


def options():
    configuration = MercuryConfiguration(
        'mercury-rpc-worker',
        BACKEND_CONFIG_FILE,
        description='Manager process for RPC workers'
    )

    add_common_options(configuration)

    configuration.add_option('backend.redis.host',
                             default='localhost',
                             help_string='Redis server address')

    configuration.add_option('backend.redis.port',
                             default=6379,
                             special_type=int,
                             help_string='Redis server port')

    configuration.add_option('backend.redis.queue',
                             default='rpc_task_queue',
                             help_string='The queue to use for RPC tasks')

    configuration.add_option('backend.workers.threads',
                             special_type=int,
                             default=4)

    configuration.add_option('backend.workers.max_requests_per_thread',
                             special_type=int,
                             default=100)

    configuration.add_option('backend.rpc_router',
                             required=True,
                             help_string='The RPC service router')

    return configuration.scan_options()


class RPCTask(RedisTask):
    def __init__(self, rpc_router_url, redis_host, redis_port, redis_queue):
        """

        :param rpc_router_url:
        :param redis_host:
        :param redis_port:
        :param redis_queue:
        """
        self.rpc_router_url = rpc_router_url
        self.rpc_router = RPCClient(self.rpc_router_url,
                                    linger=0,
                                    response_timeout=5,
                                    rcv_retry=3)
        super(RPCTask, self).__init__(redis_host, redis_port, redis_queue)

    def do(self):
        url = 'tcp://{host}:{port}'.format(**self.task)
        client = RouterReqClient(url, linger=0,
                                 response_timeout=5,
                                 rcv_retry=3)

        client.service_name = 'AgentTaskService'

        _payload = {
            'category': 'rpc',
            'method': self.task['method'],
            'args': self.task['args'],
            'kwargs': self.task['kwargs'],
            'task_id': self.task['task_id'],
            'job_id': self.task['job_id']
        }
        log.info(f'Dispatching task: {self.task}')
        response = client.transceiver(_payload)
        if response.get('error'):  # Transport Error
            err_msg = f'{self.task["mercury_id"]} has gone away while ' \
                      f'handling {self.task["task_id"]}. Transport Message: ' \
                      f'{response["message"]}'
            log.error(err_msg)
            self.rpc_router.complete_task({
                'job_id': self.task['job_id'],
                'task_id': self.task['task_id'],
                'status': 'ERROR',
                'message': err_msg,
            })
        elif response['message']['status'] != 0:
            self.rpc_router.complete_task({
                'job_id': self.task['job_id'],
                'task_id': self.task['task_id'],
                'status': 'ERROR',
                'message': f'Dispatch Error: {response["message"]}'})

        # close the socket
        client.close()


def configure_logging(config):
    logging.basicConfig(level=logging.getLevelName(config.logging.level),
                        format=config.logging.format)


def main():
    config = options()

    configure_logging(config)

    # Set this up for access from our threads

    manager = Manager(RPCTask, config.backend.workers.threads,
                      config.backend.workers.max_requests_per_thread,
                      handler_args=(config.backend.rpc_router,
                                    config.backend.redis.host,
                                    config.backend.redis.port,
                                    config.backend.redis.queue))
    manager.manage()


if __name__ == '__main__':
    main()
